from typing import Literal

class StallTracker:
    def __init__(self, name: str, relevant_thread: int | Literal["gil"]): ...
    def go_active(self) -> None: ...
    def go_idle(self) -> None: ...
    def is_active(self) -> bool: ...
    def counter_address(self) -> int: ...
    def close(self) -> None: ...
    def _leak(self) -> None: ...

def stall_gil(seconds: float) -> None: ...
